/*
==============================================================================

	RpC

==============================================================================
                                      Copyright(C)1998 Fummy (T.Ohtake)
                                      All rights reserved.
==============================================================================

NXꗗ


* End of class list
==============================================================================
$Header:

$Log: $
==============================================================================
*/

#include "secl.h"
#include <fstream.h>
#include <stdio.h>

struct { const char *name; int init; } vars[] = {
	{ "_Time", 0 },
	{ "_Delta", 48 },
	{ "_ActivePage", 0 },
	{ "_MaskPattern", 255 },
	{ "_Color", 0 },
	{ "_BackColor", 0 },
	{ "_TextWindowX1", 0 },
	{ "_TextWindowY1", 0 },
	{ "_TextWindowX2", 0 },
	{ "_TextWindowY2", 0 },
	{ "_TextPage", 0 },
	{ "_TextColor", 0 },
	{ "_TextCursorX", 0 },
	{ "_TextCursorY", 0 },
	{ "_TextPattern", 0 },
	{ "_TextMask", 255 },
	{ "_TextMode", 1 },
	{ 0, 0 }
};

Compiler::Compiler(const char *in, const char *out, const char *smf, bool db)
{
	debug = db;
	string s = "Reading ";
	s += in;
	s += "...";
	DisplayMessage(s);
	infile = new SourceFiles(in);

	output = new SBinaryOut(out);

	// LbV
	irecte = false;

	// ϐ
	variable = new IdentifyTable(0);

	for(int i = 0; vars[i].name; i++)
	{
		Variable *v = new Variable(vars[i].name, vars[i].init);
		variable->Add(v);
	}

	variable->Add(new Variable("_DisplayPage", Number, MAX_DISPPAGE, -1));

	// ֐\
	// function = new IdentifyTable(0);

	// SMF֘A
	smftimebase = 192;
	tms = 0;
	endtime = 0;

	s = "Read SMF ";
	s += smf;
	s += "...";
	if (!ReadSMF(smf))
		s += "failed. ignored time signature information.";
	else
		s += "ok.";
	DisplayMessage(s);
}

Compiler::~Compiler()
{
	delete output;
	delete variable;
//	delete function;
}

#define WORD_LE(p) ( (p)[0] << 8 | (p)[1] )

int GetV(istream &in)
{
	unsigned long n = 0;

	for(int i = 0;; i++)
	{
		if (in.eof())
			break;

		int c = in.get();

		n = (n << 7) | (unsigned long)(c & 0x7FU);

		if (!(c & 0x80U))
		{
			break;
		}
	}
	return n;
}

bool Compiler::ReadSMF(const char *fname)
{
	ifstream in(fname, ios::binary);
	if (!in)
		return false;

	char buff[256];

	in.read(buff, 4);
	if (memcmp(buff, "MThd", 4))
		return false;

	in.read(buff, 4);

	// format type;
	in.read(buff, 2);

	// tracks
	in.read(buff, 2);
	int tracks = WORD_LE(buff);

	// time base
	in.read(buff, 2);
	smftimebase = WORD_LE(buff);
	int measuresize = smftimebase*4;

	for(int i = 0; i < tracks; i++)
	{
	// MTrk
	in.read(buff, 4);
	if (memcmp(buff, "MTrk", 4))
		return false;

	// track length
	in.read(buff, 4);

	int running = 0x80;
	int count = 0;
	int measure = 0;
	int tick = 0;

	while(!in.eof())
	{
		int delta = GetV(in);

		count += delta;

		if (!i)
		{
			tick += delta;

			if (tick >= measuresize)
			{
				measure += tick/measuresize;
				tick %= measuresize;
			}
		}

		int cmd = in.peek();
		if (cmd & 0x80)
		{
			in.get();
			if (cmd < 0xF0)
				running = cmd;
		}
		else
			cmd = running;

		int cmdf = cmd & 0xF0;

		if (cmd == 0xFF)
		{
			// meta
			int meta = in.get();
			int len = GetV(in);
			char *buff = new char[len];
			in.read(buff, len);

			if (meta == 0x2F)
			{
				break;
			}
			else if (meta == 0x58 && i == 0)
			{
				int bpm = buff[0] ? buff[0] : 4;
				int bs = 1 << buff[1];

				measuresize = bpm*smftimebase*4/bs;

				tm[tms].measure = measure;
				tm[tms].tick = count;
				tm[tms].measuresize = measuresize;

				tms++;
			}

			delete[] buff;
		}
		else if (cmd == 0xF0 || cmd == 0xF7)
		{
			// SysEx
			int len = GetV(in);
			in.seekg(len, ios::cur);
		}
		else if (cmdf == 0xD0 || cmdf == 0xC0)
		{
			in.get();
		}
		else
		{
			in.get();
			in.get();
		}
	}

	if (endtime < count)
		endtime = count;
	}
	if (endtime)
		return true;
	else
		return false;
}

int Compiler::GetTick(int meas)
{
	int t = 0;
	int msz = smftimebase*4;

	for(int i = 0, j = 0; i < meas;)
	{
		if (j < tms && tm[j].measure == i)
		{
			msz = tm[j].measuresize;
			j++;
		}
		if (j >= tms)
		{
			t += msz*(meas-i);
			break;
		}

		t += msz*(tm[j].measure-i);
		i = tm[j].measure;
	}

	return t;
}

string Compiler::ConvertPicture(const char *fname)
{
	const char *ext = GetExt(fname);

	if (!stricmp(ext, "png"))
		return fname;
	else if (!stricmp(ext, "bmp"))
	{
		char b[256];
		string s = "Convert from BMP to PNG '";
		s += fname;
		s += "'";
		DisplayMessage(s);
		sprintf(b, "bmp2png %s", fname);
		system(b);

		return AddExt(fname, "png");
	}
	else if (!stricmp(ext, "gif"))
	{
		char b[256];
		string s = "Convert from GIF to PNG '";
		s += fname;
		s += "'";
		DisplayMessage(s);
		sprintf(b, "gif2png %s", fname);
		system(b);

		return AddExt(fname, "png");
	}
	else
	{
		return "";
	}
}

void Compiler::DisplayMessage(const char *s)
{
	printf("\n");
	printf(s);
}

void Compiler::ErrorReport(Error *err)
{
	const char *kk;

	if (err->type == Error::error)
		kk = "Error: ";
	else if (err->type == Error::warnning)
		kk = "Warnning: ";
	else if (err->type == Error::fatal)
		kk = "Fatal error: ";

	char b[1024];
	if (err->fname)
		sprintf(b, "%s%s (%d): %s", kk, err->fname, err->lineno, err->msg.c_str());
	else
		sprintf(b, "%s%s", kk, err->msg.c_str());

	DisplayMessage(b);

	if (err->line)
	{
		string s;

		const char *p = err->line;
		s = "->";
		while(*p != '\n' && *p != '\r' && *p)
		{
			char b[3];
			b[0] = *p++;
			b[1] = 0;

			s += b;
		}

		DisplayMessage(s);

		s = "  ";
		for(int i = 0; i < err->point; i++)
			s += " ";
		s += "^";
		DisplayMessage(s);
	}
}
