
#include "swrdout.h"
#include <cstring.h>
#include "ver.h"

#define COMPILER_NAME "Sherry WRD Easy Compiler ver." VERDESC

#define TRANSFER_VRAM 65535
#define GTEXTPAT_VRAM 65534
#define TEXT_VRAM     65533
#define WORK_VRAM     65532
#define SMOOTH_VRAM   65531

#define DUMMY_PALETTE 65535			// DrawPictureł悤
#define WORK_PALETTE  65534

#define MAX_PAGE 32767
#define MAX_PALETTE 32767

#define MAX_DISPPAGE 4
#define MAX_TIMESIGNATURE 256

string AddExt(const char *fname, const char *ext);
const char *GetExt(const char *fname);

class Compiler;


struct TimeSignature
{
	int measure;
	int tick;
	int measuresize;
};

class Error
{
	public:
	string msg;
	const char *fname;
	const char *line;
	int lineno;
	int point;

	enum ErrType {error, fatal, warnning} type;

	Error() {line = 0; point = 0;}
};

/*
class CompilerUserHandler
{
	public:
	virtual void ErrorHandler(const Error *) {}
	virtual void CompileReport(int line, int per, );
};
*/

// ==================================================================
// ʎq
// ==================================================================
struct Identify
{
	string name;

	Identify *next;

	virtual ~Identify() {}
	Identify(const char *p) {name=p;}
	Identify() {}
};

#define IDT_HASH_SIZE 64

class IdentifyTable
{
	Identify *table[IDT_HASH_SIZE];
	IdentifyTable *parent;

	Identify *SearchImp(const char *name);

	public:
	IdentifyTable(IdentifyTable *parent);
	virtual ~IdentifyTable();

	Identify *Search(const char *name);
	bool Add(Identify *);
};

// ==================================================================
// ֐^ϐǗ
// ==================================================================
enum VariableType { VarString, VarNumber, VarArray, VarNull };

struct Variable : public Identify
{
	string str;
	double number;			// Array̎Azvf
	Variable *array;

	VariableType type;

	Variable(const char *name, VariableType t) : Identify(name)
	 { number = 0; array = 0; type = t; }

	Variable(const char *name, int n = 0) : Identify(name)
	 {number = n; array = 0; type = VarNumber;}

	Variable(const char *name, VariableType t, int arrays, int init = 0);

	Variable(const char *name, const char *s) : Identify(name)
	 { str = s; type = VarString; array = 0; }

	Variable() {type = VarNull; array = 0; number = 0;}

	virtual ~Variable() {delete[] array;}

	Variable *GetElement(int n)
	 {return (n >= 0 && n < number) ? array+n : (Variable *)0;}
};

struct Function : public Identify
{
	int locate;		// ֐̈ʒu

	int vars;
	Variable *variablelist;

	Function() {}
	~Function() {}
};

struct SystemFunction : public Identify
{
	void (*system)(int, Variable *);
};

// ==================================================================
//
// ==================================================================

// LVJAiCU

enum TokenType {
	Null,
	Number,			// l '0'-'9' or '0x' '0'-'F' or '0' '0'-'8'
	String,			//  "..."
	IdentifyName,	// ʎq
	Text,			// eLXgp
	Plus,			// '+'
	Minus,			// '-'
	Eval,			// '='
	Equal,			// '=='
	Lt,				// '<'
	Gt,				// '>'
	LtEq,			// '<='
	GtEq,			// '>='
	NotEqual,		// '!='
	Multiply,		// '*'
	Division,		// '/'
	Mod,			// '%'
	And,			// '&'
	Not,			// '!'
	Or,				// '|'
	Comma,			// ','
	Semicolon,		// ';'
	Colon,			// ':'
	TextStart,		// '$'
	Open,
	Close,
	OpenBlock,
	CloseBlock,
};

struct Token
{
	enum TokenType type;
	int priority;

	int number;
	string str;		// /ʎq̂

	int locate;		// t@C̈ʒu

	Token *next, *prev;
};

// ==================================================================
// vvZbT & LVJAiCU
// ==================================================================

#define MAX_SOURCEFILE 100

struct Line
{
	int source;
	int line;
	int pointer;

	Line *next;
};

class SourceFiles
{
	string source[MAX_SOURCEFILE];
	int sources;
	Line *linetail;
	Line *linetop;

	char *buffer;
	int bufflen;

	char *pp;
	const char *p;

	Token *next;

	void AddFile(const char *fname);
	void ExpandBuffer(int len);

	void PreProcess(const char *);

	public:
	SourceFiles(const char *fname);
	~SourceFiles();

	const char *GetSource(void) {return buffer;}
	int GetSources(void) {return sources;}
	const char *GetSourceFilename(int i) {return source[i].c_str();}
	const Line &GetCurrentLine() {return GetLineData(p-buffer);}
	const Line &GetLineData(int point);

	void SetLocate(int locate) {p = buffer+locate;}
	int GetLocate(void) {return (int)(p-buffer);}

	Token *GetToken(void);
	Token *PeekToken(void);

	void ErrorMessage(const char *errmes, Error::ErrType et = Error::error, int locate = -1);
};

// ==================================================================
// RpC{
// ==================================================================

struct TextWindow
{
	TextWindow() {
		left = 0; top = 0; right=65535; bottom=65535;
		mask = 255; page = 0; x = 0; y = 0; color = 0; bcolor = 0;
		height = 16; width = 8;
	}

	int left;
	int top;
	int right;
	int bottom;

	int height;
	int width;
	int mask;
	int page;

	int x;
	int y;

	int color;		// -1 = eLXgp^[
	int bcolor;
};

class Compiler
{
	protected:
	// SMF֘A
	TimeSignature tm[MAX_TIMESIGNATURE];
	int tms;
	int smftimebase;
	int endtime;

	int GetTick(int measure);

	// p[T
	int Parse(IdentifyTable *var);
	Token *next;					// ǂ݃g[N

	// \[Xt@C
	SourceFiles *infile;

	// gbvx̕ϐ\/֐\
	IdentifyTable *variable;

	// ̂T[rX֐
	bool IsDisplayPage(int page);

	// @SherryWRD߂ʉ߂ǂ̃tO
	bool swrddone;

	// WRD̃^Cx[Xl
	int timebase;

	// WRD̃ItZbg
	int offset;

	// VRAMLbV֘A
	SRect irect;
	bool irecte;
	void Invalidate(const SRect &r);
	void PaintVram();

	// ⏕֐
	void DrawText(int, int , int, int, int , int, const char *);
	int DrawLyrics(int page, int tm, int x, int y, int mk, int col, const char *text, bool sm = false);
	void TextLyricsOut(const char *s, TextWindow &);
	void UseText();

	// o̓t@C
	SBinaryOut *output;

	// G[
	int errors;
	bool debug;

	// ֐p[^`FbN
	int GetNum(Token *token, double def = -1, double low = -1, double high = -1);
	const char *GetStr(Token *token);
	void CheckEnd(Token *);

	// z֐
	virtual void DisplayMessage(const char *s);
	virtual void DisplayMessage(const string &s) {DisplayMessage(s.c_str());}
	virtual void ErrorReport(Error *err);

	// VXeϐ (̂)
	int pat;		// TextPattern

	TextWindow tw;

	int time;
	int oldtime;
	int delta;

	int ton;

	int activepage;
	int mask;
	int displaypage[MAX_DISPPAGE];

	public:
	Compiler(const char *infile, const char *outfile, const char *smffile, bool debug = false);
	~Compiler();

	int Execute(void);
	void Output();
	bool ReadSMF(const char *fname);
	string ConvertPicture(const char *fname);

	// VXe֐
	public:
	void _SherryWRD(Token *top);
	void _AllocatePage(Token *top);
	void _FreePage(Token *top);
	void _FreePalatte(Token *top);
	void _AllocatePalette(Token *top);
	void _LoadPicture(Token *top);
	void _DrawPicture(Token *top);
	void _DrawPictureTransparent(Token *top);
	void _DisplayPage(Token *top);
	void _ActivePage(Token *top);
	void _MaskPattern(Token *top);
	void _Line(Token *top);
	void _BoxFill(Token *top);
	void _Box(Token *top);
	void _Fade(Token *top);
	void _PaletteMix(Token *top);
	void _PaletteContrast(Token *top);
	void _MoveEffect(Token *top);
	void _MoveTransparent(Token *top);
	void _Move(Token *top);
	void _ChangePalette(Token *top);
	void _SetPaletteAll(Token *top);
	void _SetPalette(Token *top);
	void _Start(Token *top);
	void _SetTime(Token *top);
	void _SetDelta(Token *top);
	void _Wait(Token *top);
	void _TextWindow(Token *top);
	void _TextLocate(Token *top);
	void _TextColor(Token *top);
	void _TextMask(Token *top);
	void _TextPage(Token *top);
	void _TextMode(Token *top);
	void _TextClear(Token *top);
	void _TextPatternOFF(Token *top);
	void _TextPattern(Token *top);
	void _TextHome(Token *top);
	void _DrawText(Token *top);
	void _DrawLyricsRainbow(Token *top);
	void _DrawLyrics(Token *top);
	void _CopyPalette(Token *top);
	void _End(Token *top);

};

