/*
==============================================================================

	SECL vvZbTALVJAiCU

==============================================================================
                                      Copyright(C)1998 Fummy (T.Ohtake)
                                      All rights reserved.
==============================================================================

NXꗗ


* End of class list
==============================================================================
$Header:

$Log: $
==============================================================================
*/

#include "secl.h"
#include <jctype.h>
#include <stdio.h>

// ==================================================================
// SourceFilesNX
// ==================================================================

SourceFiles::SourceFiles(const char *fname)
{
	buffer = 0;
	bufflen = 0;
	pp = 0;
	sources = 0;
	linetail = linetop = 0;

	AddFile(fname);

	p = buffer;
}

void SourceFiles::AddFile(const char *fname)
{
	if (sources >= MAX_SOURCEFILE)
		ErrorMessage("#includet@C܂");

	FILE *fp = fopen(fname, "r");

	if (!fp)
	{
		char b[1024];
		sprintf(b, "t@C܂ '%s'", fname);
		ErrorMessage(b);
	}

	source[sources] = fname;
	int s = sources;
	sources++;

	fseek(fp, 0, SEEK_END);

	int len = ftell(fp);
	fseek(fp, 0, SEEK_SET);

	ExpandBuffer(bufflen+len);
	char linebuffer[2048];

	if (!pp)
		pp = buffer;

	int line = 1;

	for(;;)
	{
		if (!fgets(linebuffer, 2048, fp))
			break;

		const char *q = linebuffer;

		while(isspace(*q))
			q++;

		if (*q == '#')
		{
			// vvZbT
			PreProcess(q);
		}
		else
		{
			int i = strlen(linebuffer);

			memcpy(pp, linebuffer, i);

			Line *ln = new Line;
			ln->source = s;
			ln->line = line++;
			ln->pointer = (int)(pp-buffer);
			if (linetail)
			{
				linetail->next = ln;
				linetail = ln;
				ln->next = 0;
			}
			else
			{
				linetail = ln;
				ln->next = 0;
			}
			if (!linetop)
				linetop = linetail;

			pp += i;
		}
	}

	fclose(fp);
}

void SourceFiles::PreProcess(const char *q)
{
	if (!strnicmp(q, "#include", 8))
	{
		q += 8;

		while(isspace(*q))
			q++;

		if (*q != 0x22)
		{
			ErrorMessage("#include̍\Ⴂ܂B");
			return;
		}
		q++;
		const char *qq = q;
		while(*q != 0x22 && *q)
			q++;

		if (*q != 0x22)
		{
			ErrorMessage("#include̍\Ⴂ܂B");
			return;
		}

		string fname = string(qq, 0, q-qq);
		AddFile(fname.c_str());
	}
}

void SourceFiles::ExpandBuffer(int len)
{
	if (bufflen >= len)
		return;

	char *nb = new char[len];
	bufflen = len;
	if (buffer)
	{
		memcpy(nb, buffer, bufflen);
		delete[] buffer;
	}
	buffer = nb;
}

SourceFiles::~SourceFiles()
{
}

void SourceFiles::ErrorMessage(const char *str, Error::ErrType et, int locate)
{
	Error *err = new Error;

	if (locate < 0)
		locate = GetLocate();

	const Line *l = &GetLineData(locate);

	err->msg = str;
	err->line = l ? buffer+l->pointer : 0;
	err->lineno = l ? l->line : 0;
	err->point = l ? locate-l->pointer : 0;
	err->type = et;
	err->fname = l ? source[l->source].c_str() : 0;

	throw err;
}

const Line &SourceFiles::GetLineData(int p)
{
	Line *l;

	for(l = linetop; l; l = l->next)
	{
		if (p < l->pointer || !l->next)
			break;
	}

	return *l;
}

struct
{
	enum TokenType type;
	const char *str;
	int priority;
} lextable[] = {
	{ Plus, "+" , 3},
	{ Minus, "-" , 3},
	{ Equal, "==" , 5},
	{ Eval, "=" , 8},
	{ LtEq, "<=" , 4},
	{ GtEq, ">=" , 4},
	{ Lt, "<" , 4},
	{ Gt, ">" , 4},
	{ NotEqual, "!=" , 5},
	{ Multiply, "*", 2 },
	{ Division, "/" , 2 },
	{ Mod, "%" , 2},
	{ And, "&" , 6},
	{ Not, "!" , 1},
	{ TextStart, "$" , 1},
	{ Or, "|" , 7},
	{ Comma, "," },
	{ Colon, ":" },
	{ Semicolon, ";" },
	{ Open, "(" , 0},
	{ Close, ")" , 0},
	{ OpenBlock, "{" , 0},
	{ CloseBlock, "}" , 0},
	{ Null, 0 }
};

const char *del = "+-=<>!*/%&|,:;(){}";

Token *SourceFiles::GetToken(void)
{
	delete next;
	next = 0;
	next = PeekToken();

	return next;
}

Token *SourceFiles::PeekToken(void)
{
	if (next)
		return next;

	if (!*p)
		return 0;

	Token tt;
	Token *t = &tt;
	t->locate = GetLocate();
	t->type = Null;

	while(*p)
	{
		if (isalpha(*p) || *p == '_' || *p == '@')
		{
			const char *q = p;

			while(isalpha(*p) || *p == '_' || *p == '@' || (*p >= '0' && *p <= '	9'))
			{
				const char *q;
				for(q = del; *q; q++)
					if (*q == *p)
						break;
				if (*q)
					break;

				p++;
			}

			// ϐ
			t->str = string(q, 0, p-q);
			t->type = IdentifyName;
			break;
		}
		else if (p[0] == '/' && p[1] == '/')
		{
			// Rg
			while(*p && *p != '\n')
				p++;
		}
		else if (p[0] == '\n')
		{
			p++;
		}
		else if (p[0] == '$')
		{
			p++;
			const char *q = p;
			while(*p != '\n' && *p)
			{
				if (iskanji(*p))
					p++;
				p++;
			}

			t->str = string(q, 0, p-q);
			t->type = TextStart;

			if (*p == '\n')
				p++;
			break;
		}
		else if (p[0] == ' ' || p[0] == '\t')
		{
			while(*p == ' ' || *p == '\t')
				p++;
		}
		else if (p[0] == '0' && (p[1] == 'x' || p[1] == 'X'))
		{
			// 16i
			const char *q = p;
			p++; p++;
			while((*p >= 'A' && *p <= 'F') || (*p >= 'a' && *p <= 'f') || (*p >= '0' && *p <= '9'))
					p++;

				t->number = strtol(q, 0, 0);
				t->type = Number;
				break;
		}
		else if (*p >= '1' && *p <= '9')
		{
			// 10i
			const char *q = p;
			p++;

			while((*p >= 'A' && *p <= 'F') || (*p >= 'a' && *p <= 'f') || (*p >= '0' && *p <= '9'))
				p++;

			t->number = strtol(q, 0, 0);
			t->type = Number;
			break;
		}
		else if (*p == '0')
		{
			// 8i
			const char *q = p;
			p++;

			while((*p >= 'A' && *p <= 'F') || (*p >= 'a' && *p <= 'f') || (*p >= '0' && *p <= '9'))
				p++;

			t->number = strtol(q, 0, 0);
			t->type = Number;
			break;
		}
		else if (*p == '$')
		{
			p++;
			const char *q = p;
			while(*p != '\n' && *p)
				p++;

			t->str = string(q, 0, p-q);
			t->type = Text;
			break;
		}
		else if (*p == '\x22')
		{
			int len;
			//const char *q = p+1;

			char bf[3];
			bf[1] = 0;

			p++;
			while(*p != '\x22')
			{
				if (iskanji(*p))
				{
					bf[0] = *p++;
					if (*p && *p != '\n')
						bf[1] = *p++;
					bf[2] = 0;
					t->str += bf;
				}
				else if (*p == '\\')
				{
					p++;
					if (*p == '\n')
					{
						char b[256];
						sprintf(b, "pĂ܂");
						ErrorMessage(b);
					}
					bf[0] = *p++;
					t->str += bf;
				}
				else if (!*p || *p == '\n')
				{
					char b[256];
					sprintf(b, "pĂ܂");
					ErrorMessage(b);
				}
				else
				{
					bf[0] = *p++;
					bf[1] = 0;
					t->str += bf;
				}
			}
			p++;
			t->type = String;
			break;
		}
		else
		{
			int i;

			for(i = 0; lextable[i].str; i++)
			{
				if (!strncmp(lextable[i].str, p, strlen(lextable[i].str)))
				{
					t->type = lextable[i].type;
					t->str = lextable[i].str;
					p += strlen(lextable[i].str);
					break;
				}
			}
			if (!lextable[i].str)
			{
				char b[256];
				sprintf(b, "@G[");
				ErrorMessage(b);
			}
			else
				break;
		}
	}

	if (t->type == Null)
	{
		return 0;
	}

	t = new Token;
	*t = tt;
	next = t;
	return t;
}


