
#include "swrdout.h"
#include <fstream.h>

SOutBuffer::SOutBuffer(int cmd)
 : command(cmd)
{
	buffer = 0;
	bufflen = 0;
	datasize = 0;
}

SOutBuffer::~SOutBuffer()
{
	delete[] buffer;
}

void SOutBuffer::Alloc(int sz)
{
	int ns = sz+datasize;

	if (ns >= bufflen)
	{
		int d = (ns+255)&0xFFFFFF00;
		char *nb = new char[bufflen+d];
		memcpy(nb, buffer, bufflen);
		bufflen += d;
		delete[] buffer;
		buffer = nb;
	}
}

void SOutBuffer::OutByte(SByte n)
{
	Alloc(1);

	buffer[datasize++] = n;
}

void SOutBuffer::OutWord(SWord n)
{
	Alloc(2);

	buffer[datasize++] = (SByte)(n >> 0);
	buffer[datasize++] = (SByte)(n >> 8);
}

void SOutBuffer::OutDword(SDword n)
{
	Alloc(4);

	buffer[datasize++] = (SByte)(n >> 0);
	buffer[datasize++] = (SByte)(n >> 8);
	buffer[datasize++] = (SByte)(n >> 16);
	buffer[datasize++] = (SByte)(n >> 24);
}

void SOutBuffer::OutVal(SDword val)
{
	long buffer;

	buffer = val & 0x7F;
	while ((val >>= 7) > 0)
	{
		buffer <<= 8;
		buffer |= 0x80;
		buffer |= (val & 0x7F);
	}

	while(true)
	{
		OutByte(buffer & 0xFF);
		if (buffer & 0x80)
			buffer >>= 8;
		else
			break;
	}
}

void SOutBuffer::OutArray(const char *data, int len)
{
	if (!len)
		return;

	Alloc(len);

	memcpy(buffer+datasize, data, len);
	datasize += len;
}

SBinaryOut::SBinaryOut(const char *filename)
{
	fileown = true;
	file = new ofstream(filename, ios::binary | ios::out);
	Header();
	beforetime = time = 0;
	packet = 0;
	packet_buffsize = packets = 0;
	pcount = 0;
}

SBinaryOut::SBinaryOut(ostream &o)
{
	file = &o;
	fileown = false;
	Header();
	beforetime = time = 0;
	packet = 0;
	packet_buffsize = packets = 0;
	pcount = 0;
}

SBinaryOut::~SBinaryOut()
{
	if (phase == HeaderBlock)
		End();
	if (phase == DataBlock)
		End();

	if (fileown)
		delete file;

	FreePackets();
}

void SBinaryOut::swap_int(Packet **a, Packet **b)
{
	Packet *t = *b;
	*b = *a;
	*a = t;
}

void SBinaryOut::qsort_tick(Packet **v, int left, int right)
{
	if (left >= right)
		return;

	int i, last;

	swap_int(&v[left], &v[(left+right)/2]);
	last = left;
	for(i = left+1; i <= right; i++)
	{
		if ((v[i]->tick < v[left]->tick) ||
		    (v[i]->tick == v[left]->tick && v[i]->count < v[left]->count))
			swap_int(&v[++last], &v[i]);
	}
	swap_int(&v[left], &v[last]);
	qsort_tick(v, left, last-1);
	qsort_tick(v, last+1, right);
}

void SBinaryOut::Sorting()
{
	qsort_tick(packet, 0, packets-1);
}

void SBinaryOut::OutPacket(SOutBuffer *b)
{
	if (packets >= packet_buffsize)
	{
		Packet **b = new Packet *[packet_buffsize+2048];
		int i;
		for(i = 0; i < packet_buffsize; i++)
		{
			b[i] = packet[i];
		}

		packet_buffsize += 2048;

		for(; i < packet_buffsize; i++)
			b[i] = 0;

		delete[] packet;
		packet = b;
	}

	packet[packets++] = new Packet(time, pcount, b);
	pcount ++;
}

void SBinaryOut::FreePackets()
{
	for(int i = 0; i < packet_buffsize; i++)
		delete packet[i];

	delete[] packet;
	packet = 0;
	packet_buffsize = 0;
	packets = 0;
	pcount = 0;
}
void SBinaryOut::OutPackets()
{
	SDword t = 0;

	for(int i = 0; i < packets; i++)
	{
		if (phase == DataBlock)
		{
			SDword nt = packet[i]->tick;

			OutVal(nt-t);
			t = nt;
		}

		OutVal(packet[i]->buffer->GetLength()+1);
		OutByte(packet[i]->buffer->GetCommand());
		OutBuffer(*packet[i]->buffer);
	}
}

void SBinaryOut::Header()
{
	OutArray("Sherry WRD\0\0", 12);
	OutDword(1);

	phase = HeaderBlock;
}

void SBinaryOut::OutVal(SDword val)
{
	long buffer;

	buffer = val & 0x7F;
	while ((val >>= 7) > 0)
	{
		buffer <<= 8;
		buffer |= 0x80;
		buffer += (val & 0x7F);
	}

	while(true)
	{
		OutByte(buffer& 0xFF);
		if (buffer & 0x80)
			buffer >>= 8;
		else
			break;
	}
}

void SBinaryOut::OutByte(SByte n)
{
	file->put((char)(n));
}

void SBinaryOut::OutWord(SWord n)
{
	file->put((char)((n >> 0) & 0xFF));
	file->put((char)((n >> 8) & 0xFF));
}

void SBinaryOut::OutDword(SDword n)
{
	file->put((char)((n >> 0) & 0xFF));
	file->put((char)((n >> 8) & 0xFF));
	file->put((char)((n >> 16) & 0xFF));
	file->put((char)((n >> 24) & 0xFF));
}

void SBinaryOut::OutArray(const char *dt, int len)
{
	file->write(dt, len);
}

void SBinaryOut::OutString(const char *dt)
{
	file->write(dt, strlen(dt));
}

void SBinaryOut::CheckHeader()
{
	if (phase == HeaderBlock)
		return;

	throw new SError("Cannot define command in Data block");
}

void SBinaryOut::CheckData()
{
	if (phase != DataBlock)
		throw new SError("Must be WRD command in Data block");
}

void SBinaryOut::OutDelta()
{
	//CheckDataDelta();
	OutVal(time-beforetime);
	beforetime = time;
}

void SBinaryOut::SetTime(SDword t)
{
	time = t;
}

void SBinaryOut::AddTime(SDword t)
{
	time += t;
}

void SBinaryOut::End()
{
	Sorting();

	if (packets)
		SetTime(packet[packets-1]->tick);
	SOutBuffer *buff = new SOutBuffer(0);
	OutPacket(buff);

	OutPackets();
	FreePackets();

	if (phase == HeaderBlock)
		phase = DataBlock;
	else if (phase == DataBlock)
		phase = Done;
}

void SBinaryOut::TimeBase(int tb)
{
	SOutBuffer *buff = new SOutBuffer(0x21);
	buff->OutWord(tb);

	OutPacket(buff);
}

void SBinaryOut::CompilerName(const char *str)
{
	SOutBuffer *buff = new SOutBuffer(0x71);
	buff->OutString(str);

	OutPacket(buff);
}

void SBinaryOut::SourceFilename(int n, const char *str)
{
	SOutBuffer *buff = new SOutBuffer(0x72);

	buff->OutWord(n);
	buff->OutString(str);

	OutPacket(buff);
}

void SBinaryOut::Start(void)
{
	SOutBuffer *buff = new SOutBuffer(0x01);
	OutPacket(buff);

	Sorting();
	OutPackets();
	FreePackets();

	time = 0;
}

void SBinaryOut::AllocatePalette(int pal)
{
	SOutBuffer *buff = new SOutBuffer(0x21);
	buff->OutWord(pal);
	OutPacket(buff);
}

void SBinaryOut::AllocatePalette(int pal[], int pals)
{
	SOutBuffer *buff = new SOutBuffer(0x21);
	for(int i = 0; i < pals; i++)
	{
		buff->OutWord(pal[i]);
	}
	OutPacket(buff);
}

void SBinaryOut::FreePalette(int pal)
{
	SOutBuffer *buff = new SOutBuffer(0x22);
	buff->OutWord(pal);
	OutPacket(buff);
}

void SBinaryOut::FreePalette(int pal[], int pals)
{
	SOutBuffer *buff = new SOutBuffer(0x22);
	for(int i = 0; i < pals; i++)
	{
		buff->OutWord(pal[i]);
	}
	OutPacket(buff);
}

void SBinaryOut::FreeScreen(int pal)
{
	SOutBuffer *buff = new SOutBuffer(0x26);
	buff->OutWord(pal);
	OutPacket(buff);
}

void SBinaryOut::FreeScreen(int pal[], int pals)
{
	SOutBuffer *buff = new SOutBuffer(0x26);
	for(int i = 0; i < pals; i++)
	{
		buff->OutWord(pal[i]);
	}
	OutPacket(buff);
}

void SBinaryOut::AllocateScreen(const VScreenDescription &vsd)
{
	SOutBuffer *buff = new SOutBuffer(0x25);

	buff->OutWord(vsd.number);
	buff->OutWord(vsd.size.x);
	buff->OutWord(vsd.size.y);
	buff->OutByte(vsd.transparent);
	OutPacket(buff);
}

void SBinaryOut::AllocateScreen(const VScreenDescription vsd[], int vsds)
{
	SOutBuffer *buff = new SOutBuffer(0x25);

	for(int i = 0; i < vsds; i++)
	{
		buff->OutWord(vsd[i].number);
		buff->OutWord(vsd[i].size.x);
		buff->OutWord(vsd[i].size.y);
		buff->OutByte(vsd[i].transparent);
	}
	OutPacket(buff);
}

void SBinaryOut::AllocateScreenWithLoad(int scr, int pal, const char *str)
{
	SOutBuffer *buff = new SOutBuffer(0x27);

	buff->OutWord(scr);
	buff->OutWord(pal);
	buff->OutString(str);
	OutPacket(buff);
}

void SBinaryOut::StorePalette(int pal)
{
	SOutBuffer *buff = new SOutBuffer(0x31);

	buff->OutWord(pal);
	OutPacket(buff);
}

void SBinaryOut::StoreVRAM(int scr, const SPoint &p)
{
	SOutBuffer *buff = new SOutBuffer(0x35);

	buff->OutWord(scr);
	buff->OutWord(p.x);
	buff->OutWord(p.y);
	OutPacket(buff);
}

void SBinaryOut::StoreVRAM(int scr, const SPoint &p, const SPoint &p2, const SPoint &p3)
{
	SOutBuffer *buff = new SOutBuffer(0x36);

	buff->OutWord(scr);
	buff->OutWord(p.x);
	buff->OutWord(p.y);
	buff->OutWord(p2.x);
	buff->OutWord(p2.y);
	buff->OutWord(p3.x);
	buff->OutWord(p3.y);
	OutPacket(buff);
}

void SBinaryOut::SetPalette(int palnum, const SBinaryOut::Palette pal[], int palettes)
{
	SOutBuffer *buff = new SOutBuffer(0x41);

	buff->OutWord(palnum);
	buff->OutArray((const char *)pal, palettes*sizeof(Palette));
	OutPacket(buff);
}

void SBinaryOut::CopyPalette(int src, int dest)
{
	SOutBuffer *buff = new SOutBuffer(0x43);

	buff->OutWord(src);
	buff->OutWord(dest);
	OutPacket(buff);
}

void SBinaryOut::CopyMaskPalette(int src, int dest, char mask[])
{
	SOutBuffer *buff = new SOutBuffer(0x44);

	buff->OutWord(src);
	buff->OutWord(dest);
	buff->OutArray(mask, 32);
	OutPacket(buff);
}

void SBinaryOut::MergePalette(int pal1, int pal2, int palout, int pal1s, int pal1m, int pal1out, int pal2s, int pal2m, int pal2out, int per1, int per2)
{
	SOutBuffer *buff = new SOutBuffer(0x42);

	buff->OutWord(pal1);
	buff->OutWord(pal2);
	buff->OutWord(palout);
	buff->OutByte(pal1s);
	buff->OutByte(pal1m);
	buff->OutByte(pal1out);
	buff->OutByte(pal2s);
	buff->OutByte(pal2m);
	buff->OutByte(pal2out);
	buff->OutByte(per1);
	buff->OutByte(per2);
	OutPacket(buff);
}

void SBinaryOut::DrawString(int scr, int mask, int mode, int color, int bgcolor, const SPoint &p, const char *str)
{
	SOutBuffer *buff = new SOutBuffer(0x51);

	buff->OutWord(scr);
	buff->OutByte(mask);
	buff->OutByte(mode);
	buff->OutByte(color);
	buff->OutByte(bgcolor);
	buff->OutWord(p.x);
	buff->OutWord(p.y);
	buff->OutString(str);
	OutPacket(buff);
}

void SBinaryOut::DrawBoxFill(int scr, int mask, const SRect &r, int color)
{
	SOutBuffer *buff = new SOutBuffer(0x52);

	buff->OutWord(scr);
	buff->OutByte(mask);
	buff->OutWord(r.x1);
	buff->OutWord(r.y1);
	buff->OutWord(r.x2);
	buff->OutWord(r.y2);
	buff->OutByte(color);
	OutPacket(buff);
}

void SBinaryOut::DrawHLine(int scr, int mask, int x1, int x2, int y1, int color)
{
	SOutBuffer *buff = new SOutBuffer(0x54);

	buff->OutWord(scr);
	buff->OutByte(mask);
	buff->OutWord(x1);
	buff->OutWord(x2);
	buff->OutWord(y1);
	buff->OutByte(color);
	OutPacket(buff);
}

void SBinaryOut::DrawVLine(int scr, int mask, int x1, int y1, int y2, int color)
{
	SOutBuffer *buff = new SOutBuffer(0x54);

	buff->OutWord(scr);
	buff->OutByte(mask);
	buff->OutWord(x1);
	buff->OutWord(y1);
	buff->OutWord(y2);
	buff->OutByte(color);
	OutPacket(buff);
}

void SBinaryOut::DrawLine(int src, int mask, const SRect &r, int color)
{
	SOutBuffer *buff = new SOutBuffer(0x55);

	buff->OutWord(src);
	buff->OutByte(mask);
	buff->OutWord(r.x1);
	buff->OutWord(r.y1);
	buff->OutWord(r.x2);
	buff->OutWord(r.y2);
	buff->OutByte(color);
	OutPacket(buff);
}

void SBinaryOut::Move(int src, int dest, int mask, const SRect &srcr, const SPoint &destp)
{
	SOutBuffer *buff = new SOutBuffer(0x61);

	buff->OutWord(src);
	buff->OutWord(dest);
	buff->OutByte(mask);
	buff->OutByte(0);
	buff->OutRect(srcr);
	buff->OutPoint(destp);
	OutPacket(buff);
}

void SBinaryOut::MoveTransparent(int src, int dest, int mask, const SRect &srcr, const SPoint &destp)
{
	SOutBuffer *buff = new SOutBuffer(0x61);

	buff->OutWord(src);
	buff->OutWord(dest);
	buff->OutByte(mask);
	buff->OutByte(1);
	buff->OutRect(srcr);
	buff->OutPoint(destp);
	OutPacket(buff);
}

void SBinaryOut::MoveMask(int src, int dest, int mask, int trp, const SRect &srcr, const SPoint &destp, const SPoint &masksz, char mk[])
{
	SOutBuffer *buff = new SOutBuffer(0x62);

	buff->OutWord(src);
	buff->OutWord(dest);
	buff->OutByte(mask);
	buff->OutByte(trp);
	buff->OutRect(srcr);
	buff->OutPoint(destp);
	buff->OutByte(masksz.x);
	buff->OutByte(masksz.y);
	buff->OutArray(mk, (masksz.x+7)/8*masksz.y);
	OutPacket(buff);
}

void SBinaryOut::Debug(int file, int line)
{
	SOutBuffer *buff = new SOutBuffer(0x71);

	buff->OutWord(file);
	buff->OutWord(line);
	OutPacket(buff);
}

void SBinaryOut::Unique(const char *data, int length)
{
	SOutBuffer *buff = new SOutBuffer(0x7F);

	buff->OutArray(data, length);
	OutPacket(buff);
}

void SBinaryOut::DirectCommand(int p, const char *data, int length)
{
	SOutBuffer *buff = new SOutBuffer(p);

	buff->OutArray(data, length);
	OutPacket(buff);
}

