
#include "wrdsub.h"
#include "palette.h"
#include <iostream.h>
#include <string.h>

struct SOutBuffer
{
	SOutBuffer(int cmd = 0);
	~SOutBuffer();

	char *buffer;
	int bufflen;
	int datasize;

	SByte command;

	operator const char *() {return buffer;}
	int GetLength() {return datasize;}
	int GetCommand() {return command;}

	void Alloc(int len);
	void OutByte(SByte);
	void OutWord(SWord);
	void OutDword(SDword);
	void OutVal(SDword);
	void OutArray(const char *data, int len);
	void OutString(const char *str) {OutArray(str, strlen(str));}
	void OutPoint(const SPoint &p)
		{OutWord(p.x); OutWord(p.y);}
	void OutRect(const SRect &r)
		{OutWord(r.x1); OutWord(r.y1); OutWord(r.x2); OutWord(r.y2); }
};

class SBinaryOut
{
	enum {HeaderBlock, DataBlock, Done} phase;		// ݏݒ̃ubN

	struct Packet
	{
		SOutBuffer *buffer;
		SDword tick;
		SDword count;

		Packet(SDword t, SDword p, SOutBuffer *b) { buffer = b; count = p; tick = t; }
		~Packet() { delete buffer; }
	};

	Packet **packet;
	int packets;
	int packet_buffsize;
	int pcount;
	void qsort_tick(Packet **list, int left, int right);
	void swap_int(Packet **a, Packet **b);
	void Sorting();
	void FreePackets();
	void OutPacket(SOutBuffer *);
	void OutPackets();

	void OutByte(SByte data);
	void OutWord(SWord data);
	void OutDword(SDword data);
	void OutVal(SDword data);
	void OutArray(const char *data, int len);
	void OutString(const char *str);
	void OutBuffer(SOutBuffer &b) {OutArray(b, b.GetLength());}

	void Header();

	ostream *file;
	bool fileown;

	SDword beforetime;
	SDword time;

	void CheckHeader();
	void CheckData();
	void OutDelta();

	public:
	SBinaryOut(const char *fname);
	SBinaryOut(ostream &);
	~SBinaryOut();

	struct VScreenDescription
	{
		SWord number;
		SPoint size;
		SByte transparent;
	};

	struct Palette
	{
		SByte number;
		SByte red;
		SByte green;
		SByte blue;
	};

	// 
	void End();

	// wb_ubN
	void TimeBase(int);
	void CompilerName(const char *str);
	void SourceFilename(int number, const char *fname);

	// f[^ubN
	void SetTime(SDword count);
	void AddTime(SDword delta);

	void Start(void);

	void AllocatePalette(int pal[], int pals);
	void AllocatePalette(int pal);
	void FreePalette(int pal[], int pals);
	void FreePalette(int pal);
	void FreeScreen(int pal[], int pals);
	void FreeScreen(int pal);
	void AllocateScreen(const VScreenDescription [], int scrs);
	void AllocateScreen(const VScreenDescription &);
	void AllocateScreenWithLoad(int scr, int pal, const char *str);

	void StorePalette(int pal);
	void StoreVRAM(int scr, const SPoint &p);
	void StoreVRAM(int scr, const SPoint &, const SPoint &, const SPoint &sz);

	void SetPalette(int palnum, const Palette pal[], int palettes);
	void MergePalette(int pal1, int pal2, int palout, int pal1s, int pal1m, int pal1out, int pal2s, int pal2m, int pal2out, int per1, int per2);
	void CopyPalette(int src, int dest);
	void CopyMaskPalette(int src, int dest, char mask[]);

	void DrawString(int scr, int mask, int mode, int color, int bgcolor, const SPoint &, const char *str);
	void DrawBoxFill(int scr, int mask, const SRect &, int color);
	void DrawHLine(int scr, int mask, int x1, int x2, int y1, int color);
	void DrawVLine(int scr, int mask, int x1, int y1, int y2, int color);
	void DrawLine(int scr, int mask, const SRect &, int color);
	void Move(int src, int dest, int mask, const SRect &srcr, const SPoint &destp);
	void MoveTransparent(int src, int dest, int mask, const SRect &srcr, const SPoint &destp);
	void MoveMask(int src, int dest, int mask, int trp, const SRect &srcr, const SPoint &destp, const SPoint &masksz, char mk[]);
//	void MovePlane(...);

	void Debug(int file, int line);
	void Unique(const char *data, int length);
	void DirectCommand(int cmd, const char *data, int length);
};

